import os
import pickle
import logging
from dataclasses import dataclass
from typing import List, Optional
from tqdm.auto import tqdm
from filelock import FileLock

import torch
from torch.utils.data.dataset import Dataset as TorchDataset


DATA_BASE_PATH = os.environ['CAI_DATA_BASE_PATH']

logger = logging.getLogger(__name__)


@dataclass
class TokenTagExample:
    """A single set of features for an example of a token classification dataset."""

    input_ids: List[int]
    attention_mask: List[int]
    token_type_ids: Optional[List[int]] = None
    labels: Optional[List[int]] = None


class TokenTagDataset(TorchDataset):
    """A PyTorch dataset for fine-tuning a Huggingface transformer model on a token classification dataset, such as the
        SOAS part-of-speech task.

    Args:
        processed_dataset: Name of the preprocessed dataset with a pickle file with tsheg-pretokenized tokens and
            labels. Will translate to the file $CAI_DATA_BASE_PATH/{processed_dataset}/dataset.pkl.
        concatenate_examples (bool): Concatenate examples to form long sentences. Defaults to False.
        use_mask_for_word_pieces (bool): When set to False, the first piece of a word in the dataset is marked with its
            label and the rest of the pieces of that word are marked with the padding token, which is set to the
            ignored index of the cross-entropy loss by default. When set to True the [MASK] token is used for this
            instead, which is not ignored by the loss, so that the model has to learn word segmentation and token
            classification end-to-end. Defaults to False.
        return_data_as_dict: Return dictionaries instead of dataclasses. Is expected by Transformers 4 data collator.
        max_seq_length: Maximum sequence length in the examples in the dataset. Defaults to 128.
        dupe_count: Number of duplicate walks through words in the SOAS dataset to make training examples. Meant to be
            used in combination with dupe_offset. The actual number of walks will be dupe_count + 1. Defaults to 0.
        dupe_offset: Offset when duplicating the walking through words in the SOAS dataset to make training examples.
            Defaults to 3.
    """

    _pad_token_label_id = torch.nn.CrossEntropyLoss().ignore_index
    _cls_token_segment_id = 0
    _sequence_segment_id = 0
    _pad_token_segment_id = 0

    processed_dataset = None
    concatenate_examples=False
    use_mask_for_word_pieces = False
    return_data_as_dict = True
    max_seq_length = 128
    dupe_count = 0
    dupe_offset = 3

    def __init__(self,
                 tokenizer,
                 processed_dataset,
                 verbose=False,
                 tqdm=tqdm,      # pylint: disable=redefined-outer-name
                 examples=None,
                 label_to_id_map=None):
        """Initialize the dataset. This constructor will prepare examples for training with the parameters specified in
            the class, especially max_seq_length.

        Args:
            tokenizer (PreTrainedTokenizer): The tokenizer to use for getting the token ids for the various special
                tokens like [CLS], [SEP] and <pad>.
            processed_dataset: Name of the preprocessed dataset with a pickle file with tsheg-pretokenized tokens and
                labels. Will translate to the file
                $CAI_DATA_BASE_PATH/{processed_dataset}/dataset.pkl.
            verbose (bool): If True will print a progress bar using TQDM. Defaults to False.
            tqdm (TQDM module, optional): Pass in the TQDM module to use for the progress bar. Useful when running in a
                notebook. Defaults to the tqdm you get when using "from tqdm.auto import tqdm".
            examples: A list of pre-loaded examples or None. If None then will load the dataset from the preprocessed
                tokens. Defaults to None.
            label_to_id_map: A mapping of part-of-speech text labels to their tag ids. Defaults to None which means to
                infer from the dataset. Pass in an extrnal mapping if you have two separate datasets that you want to
                agree on the labels, such as separately constructed test data.
        """

        super().__init__()
        if processed_dataset.endswith(".pkl"):
            self.processed_dataset = os.path.join(DATA_BASE_PATH, f"{processed_dataset}")
        else:
            self.processed_dataset = os.path.join(DATA_BASE_PATH, f"{processed_dataset}/dataset.pkl")
        self.tokenizer = tokenizer

        if examples is not None:
            self.examples = examples
            return

        if not os.path.isfile(self.processed_dataset):
            raise FileNotFoundError("Preprocessed data not found.")

        lock_path = os.path.join(
            os.environ['CAI_TEMP_PATH'],
            os.path.basename(self.processed_dataset.replace('/', '_').replace('\\', '_')) + ".lock")
        with FileLock(lock_path):
            with open(self.processed_dataset, 'rb') as f:
                self.tokenized_data = pickle.load(f)
        concatted_data = isinstance(self.tokenized_data[0][1], list)
        examples_are_words = isinstance(self.tokenized_data[0][0][0], list)

        if concatted_data:
            unique_labels = sorted(list({label for example in self.tokenized_data for label in example[1]}))
        else:
            unique_labels = sorted(list({example[1] for example in self.tokenized_data}))
        if '[MASK]' in unique_labels:
            self.mask_token_id = unique_labels.index('[MASK]')
        else:
            self.mask_token_id = len(unique_labels) + 1
        if label_to_id_map is None:
            self.label_to_id_map = {label: idx for idx, label in enumerate(unique_labels)}
            self.label_to_id_map['<pad>'] = TokenTagDataset._pad_token_label_id
            self.label_to_id_map['[MASK]'] = self.mask_token_id
        else:
            self.label_to_id_map = label_to_id_map
            for label in set(unique_labels).difference(set(self.label_to_id_map.keys())):
                self.label_to_id_map[label] = tokenizer.unk_token_id
        self.id_to_label_map = {idx: label for label, idx in self.label_to_id_map.items()}

        self.examples, skipped_examples = [], 0
        if not self.concatenate_examples:
            for cur_tokens, cur_labels in tqdm(self.tokenized_data, desc="Example", disable=not verbose):
                cur_labels = [self.label_to_id_map[l] for l in cur_labels]
                attention_mask = [1] * len(cur_tokens)
                token_type_ids = [self._sequence_segment_id] * len(cur_tokens)
                self.examples.append(
                    TokenTagExample(
                        input_ids=cur_tokens,
                        labels=cur_labels,
                        attention_mask=attention_mask,
                        token_type_ids=token_type_ids
                    )
                )
        else:
            word_piece_id = self.mask_token_id if self.use_mask_for_word_pieces else self._pad_token_label_id
            input_ids, labels, attention_mask, token_type_ids = [tokenizer.bos_token_id], [word_piece_id], [1], \
                [self._cls_token_segment_id]
            special_tokens_count = tokenizer.num_special_tokens_to_add()

            for cur_dupe in tqdm(range(TokenTagDataset.dupe_count + 1),
                                desc="Duplicate",
                                disable=(not verbose) or (TokenTagDataset.dupe_count == 0)):
                offset = cur_dupe * TokenTagDataset.dupe_offset
                for cur_idx in tqdm(range(offset, len(self.tokenized_data)),
                                    desc="Example",
                                    disable=not verbose,
                                    leave=TokenTagDataset.dupe_count == 0):
                    cur_tokens, cur_labels = self.tokenized_data[cur_idx]
                    if examples_are_words and concatted_data:
                        new_tokens, new_labels = [], []
                        for token_batch, label in zip(cur_tokens, cur_labels):
                            new_tokens.extend(token_batch)
                            new_labels.extend([self.label_to_id_map[label]] + [word_piece_id] * (len(token_batch) - 1))
                        cur_tokens, cur_labels = new_tokens, new_labels
                    elif not concatted_data:
                        cur_labels = [self.label_to_id_map[cur_labels]] + [word_piece_id] * (len(cur_tokens) - 1)
                    if len(cur_tokens) > self.max_seq_length - special_tokens_count:
                        skipped_examples += 1
                        continue
                    new_len = len(input_ids) + len(cur_tokens)
                    if new_len >= self.max_seq_length - special_tokens_count:
                        padding_length = self.max_seq_length - special_tokens_count - len(input_ids) + 1
                        input_ids.extend([tokenizer.pad_token_id] * padding_length + [tokenizer.eos_token_id])
                        labels.extend([self._pad_token_label_id] * padding_length + [word_piece_id])
                        attention_mask.extend([0] * padding_length + [1])
                        token_type_ids.extend(
                            [self._pad_token_segment_id] * padding_length + [self._sequence_segment_id])

                        assert len(input_ids) <= self.max_seq_length
                        assert len(input_ids) == len(labels)
                        self.examples.append(
                            TokenTagExample(
                                input_ids=input_ids,
                                labels=labels,
                                attention_mask=attention_mask,
                                token_type_ids=token_type_ids))
                        input_ids, labels, attention_mask, token_type_ids = [tokenizer.bos_token_id], \
                            [word_piece_id], [1], [self._cls_token_segment_id]

                    input_ids.extend(cur_tokens)
                    labels.extend(cur_labels)
                    attention_mask.extend([1] * len(cur_tokens))
                    token_type_ids.extend([self._sequence_segment_id] * len(cur_tokens))
        if skipped_examples > 0:
            logger.warning(f"Skipped {skipped_examples} examples")

    def __len__(self):
        # Return dataset length
        return len(self.examples)

    def __getitem__(self, i):
        # Return the i-th item from the dataset
        return self.examples[i].__dict__



from .corpus_loader import CorpusSplitType
from .open_pecha_loaders import OldKangyurLoader, KangyurLoader, TengyurLoader
from .tei_loader import TeiLoader
from .parallel_tmx_loader import ParallelTMXLoader

__all__ = [
    'CorpusSplitType',
    'OldKangyurLoader',
    'KangyurLoader',
    'TengyurLoader',
    'TeiLoader',
    'ParallelTMXLoader']



from .numbers import tibetan_digits, tibetan_halves, translate_tibetan_number

try:
    from .dict import TibetanDict, TibetanEncoding
except ModuleNotFoundError:
    # For inference installations we don't currently support the data pulling utilities
    pass

__all__ = [
    "tibetan_digits",
    "tibetan_halves",
    "translate_tibetan_number",
    "TibetanEncoding",
    "TibetanDict"]



tibetan_digits = [
    "\u0F20",
    "\u0F21",
    "\u0F22",
    "\u0F23",
    "\u0F24",
    "\u0F25",
    "\u0F26",
    "\u0F27",
    "\u0F28",
    "\u0F29"]
tibetan_halves = [
    "\u0F2A",
    "\u0F2B",
    "\u0F2C",
    "\u0F2D",
    "\u0F2E",
    "\u0F2F",
    "\u0F30",
    "\u0F31",
    "\u0F32",
    "\u0F33"]


def translate_tibetan_number(tibetan_number):
    """Translate Tibetan numericals to Western numerals.

    Args:
        tibetan_number (str): Unicode Tibetan numeral. Cannot contain half-digits.

    Returns:
        int: Integer represented by the input."""
    res = ""
    for digit in tibetan_number:
        try:
            res += str(tibetan_digits.index(digit))
        except ValueError as e:
            raise ValueError(f"Supposed to be Tibetan number consisting of whole digits but contains {digit}.") from e
    return int(res)



import re
from bs4 import BeautifulSoup

from .corpus_loader import CorpusLoader
from .utils import repl_split_commas_or_kill


_strip_chars = {'!'}
_bad_chars = {
    '_', 'a', 'b', 'c', 'd', 'e', 'f', 'h', 'i', 'k', 'l', 'm', 'n', 'o', 'p', 'r', 's', 't', 'u', 'v', 'w', 'y',
    '\u200b'}


def _strip_out_pairs(soup, strip_chars, bad_chars, apply_markup, replace_with_suggested):
    pairs = []
    for sent_pair_tag in soup.find_all("tu"):
        props = sent_pair_tag.find_all("prop")
        folio, position = '', ''
        for prop_tag in props:
            if prop_tag.attrs['name'] == 'folio':
                folio = ' '.join(list(prop_tag.stripped_strings))
            elif prop_tag.attrs['name'] == 'position':
                position = ' '.join(list(prop_tag.stripped_strings))
        cur_pair = [' '.join(list(tag.stripped_strings)) for tag in sent_pair_tag.find_all("tuv")
                    if tag.get('xml:lang') in ['bo', 'en']]
        assert len(cur_pair) == 2
        tibetan, english = cur_pair
        if len(set(tibetan).intersection(bad_chars)) > 0:
            continue
        for c in strip_chars:
            tibetan = tibetan.replace(c, '')
        if apply_markup:
            tibetan = re.sub(
                r"[\[({][^\[\](){}]*?[})\]]",
                lambda x: repl_split_commas_or_kill(x, take_second_item=replace_with_suggested),
                tibetan)
        pairs.append((folio, position, tibetan, english))
    return pairs


def _process(args,            # pylint: disable=dangerous-default-value
             with_locator=True,
             strip_chars=_strip_chars,
             bad_chars=_bad_chars,
             apply_markup=True,
             replace_with_suggested=False):
    fn, fstr = args
    soup = BeautifulSoup(fstr, 'xml')
    cur_pairs = _strip_out_pairs(soup, strip_chars, bad_chars, apply_markup, replace_with_suggested)
    if with_locator:
        toh = re.findall(r"Toh_[\d-]+[a-z]?-", fn)
        if len(toh) == 0:
            raise ValueError(f"Filename has no Tohoku number: {fn}")
        toh = toh[0][4:-1]
        return [(fn, toh, folio, position, tibetan, english) for folio, position, tibetan, english in cur_pairs]
    else:
        return [(tibetan, english) for _, _, tibetan, english in cur_pairs]


class ParallelTMXLoader(CorpusLoader):
    """Loader for the parallel sentences dataset from 84,000. Source repo is
        https://github.com/84000/data-translation-memory.

    Attributes:
        data_glob: A glob for the location of the data after the prefix that locates the repo. The prefix is specified
            in the constructor. Defaults to: raw_datasets/84000-parallel-sentences/*.tmx
        strip_chars (set of characters): Set of characters to delete from the Tibetan strings.
        bad_chars (set of characters): Any Tibetan string that contains a character in this set is removed from the
            dataset and its pair is skipped.
    """

    _df_column_names = ["filename", "tohoku", "folio", "position", "tibetan", "english"]
    _df_meta = [['a'], ['a'], ['a'], ['a'], ['a'], ['a']]
    _apply_markup = True
    _replace_with_suggested = False
    _clean_bad_chars = True

    data_glob = "raw_datasets/84000-parallel-sentences/*.tmx"
    glob_exclusions = {
        "Modular_Passages-TM-Test.tmx"
    }
    strip_chars = _strip_chars
    bad_chars = _bad_chars

    def __init__(self,
                 glob_prefix=None,
                 glob_override=None):
        """Constructor for a corpus loader for the parallel sentences dataset from 84,000. The default settings are to
            apply markup, not to replace with suggestions in the markup, and to clean up bad characters.

        Args:
            glob_prefix (optional): The prefix to the current path for the location of the repo that contains this
                corpus. It will be concatenated with data_glob to get the full path. For example: ../../my_data
            glob_override (optional): Overrides the entire file path glob to specify the location of the corpus. For
                example: ../../my_data/84000/data-translation-memory/*.txt
        """

        super().__init__(glob_prefix=glob_prefix, glob_override=glob_override)

    def apply_markup(self, apply=True, replace_with_suggested=False):
        """Set whether to apply the markup for suggested replacements within the corpus.

        Args:
            apply: Markup is treated if apply=True. If apply=False then the markup is not touched at all and the
                brackets are left as is. Defaults to True.
            replace_with_suggested: If True then the suggested replacements in the markup are performed. This means
                that elements like (x,y) become y. Defaults to False.

        Returns:
            The corpus loader object so that methods can be chained in the functional style.
        """

        self._apply_markup = apply
        self._replace_with_suggested = replace_with_suggested
        return self

    def clean_bad_chars(self, clean=True):
        """Set whether to remove bad characters from the corpus. Each corpus has to specify its own bad character set.

        Args:
            clean: Remove bad characters if True. Defaults to True.

        Returns:
            The corpus loader object so that methods can be chained in the functional style.
        """

        self._clean_bad_chars = clean
        return self

    def _process_bag(self, bag, locators):
        # Prepares a bag, with or without locators as indicated. Applies markup in round and curly braces and removes
        #   bad characters, if requested.
        strip_chars, bad_chars = (self.strip_chars, self.bad_chars) if self._clean_bad_chars else (set(), set())
        bag = bag \
            .map(lambda x: _process(x,
                                    strip_chars=strip_chars,
                                    bad_chars=bad_chars,
                                    apply_markup=self._apply_markup)) \
            .flatten()
        if not locators:
            bag = bag \
                .map(lambda args: args[1:])
        return bag



from transformers.utils import logging
from transformers.integrations import TensorBoardCallback, rewrite_logs


logger = logging.get_logger(__name__)


class CAITensorboardCallback(TensorBoardCallback):
    """Our own Tensorboard callback that supports additional features:

    - Currently just outputting text as well as scalars on log
    """

    @staticmethod
    def replace_in_trainer(trainer):
        trainer.callback_handler.remove_callback(TensorBoardCallback)
        trainer.callback_handler.add_callback(CAITensorboardCallback)

    def on_log(self, args, state, control, logs=None, **kwargs):
        if not state.is_world_process_zero:
            return

        if self.tb_writer is None:
            self._init_summary_writer(args)

        if self.tb_writer is not None:
            logs = rewrite_logs(logs)
            for k, v in logs.items():
                if isinstance(v, (int, float)):
                    self.tb_writer.add_scalar(k, v, state.global_step)
                elif isinstance(v, str):
                    self.tb_writer.add_text(k, v, state.global_step)
                else:
                    logger.warning(
                        "Trainer is attempting to log a value of "
                        f'"{v}" of type {type(v)} for key "{k}" as a scalar. '
                        "This invocation of Tensorboard's writer.add_scalar() "
                        "is incorrect so we dropped this attribute."
                    )
            self.tb_writer.flush()



import os
import glob as globlib


def open_globs(globs, exclusions):
    """Open either a collection of globs or an individual glob and return the union of their filenames, with exclusions
        applied.

    Args:
        globs: Either a set/list of globs or a single glob string.
        exclusions: A collection of excluded filenames.

    Returns:
        The union of all full file paths in all the globs, with the base filenames appearing in exclusions removed.
    """
    if isinstance(globs, set):
        globs = list(globs)
    if isinstance(globs, list):
        globs = [fn for glob in globs for fn in globlib.glob(glob)]
    if isinstance(globs, str):
        globs = globlib.glob(globs)
    return list(filter(lambda fn: os.path.basename(fn) not in exclusions, globs))



from dataclasses import dataclass, fields
from transformers import TrainingArguments, Seq2SeqTrainingArguments
from transformers.trainer_utils import IntervalStrategy, SchedulerType, HubStrategy
from transformers.training_args import OptimizerNames


class HydraToHFConverterMixIn:
    _hf_base_class = None


    @classmethod
    def as_hf_training_args(cls, cfg):
        """Converts a Hydra parsed config to a dictionary that can be used to instantiate a Hugging Face base class,
        specified in the static member _hf_base_class.

        This is needed because of a bug in Hydra: it doesn't ignore fields with init=False, such as _n_gpu.
        """

        filtered_cfg = {
            field.name: cfg[field.name]
            for field in fields(cls._hf_base_class)
            if field.init
        }
        return cls._hf_base_class(**dict(filtered_cfg))     # pylint: disable=not-callable


@dataclass
class HydraTrainingArguments(TrainingArguments, HydraToHFConverterMixIn):
    """The enums in the Huggingface TrainingArguments have their values, not their members as the default setting. This
        is actually an error, for example values can be duplicated while members cannot.

    If feeding TrainingArguments directly to Hydra's ConfigStore, OmegaConf throws a validation error that the enum
    default values are set to something that isn't a member of that enum. This may seem annoying but this is, in fact,
    the correct behavior.

    This really should be fixed in the transformers library.
    """

    _hf_base_class = TrainingArguments

    evaluation_strategy: IntervalStrategy = IntervalStrategy.NO
    lr_scheduler_type: SchedulerType = SchedulerType.LINEAR
    logging_strategy: IntervalStrategy = IntervalStrategy.STEPS
    save_strategy: IntervalStrategy = IntervalStrategy.STEPS
    optim: OptimizerNames = OptimizerNames.ADAMW_HF
    hub_strategy: HubStrategy = HubStrategy.EVERY_SAVE


@dataclass
class HydraSeq2SeqTrainingArguments(Seq2SeqTrainingArguments, HydraToHFConverterMixIn):
    """The enums in the Huggingface Seq2SeqTrainingArguments have their values, not their members as the default
    setting. This is actually an error, for example values can be duplicated while members cannot.

    If feeding Seq2SeqTrainingArguments directly to Hydra's ConfigStore, OmegaConf throws a validation error that the
    enum default values are set to something that isn't a member of that enum. This may seem annoying but this is, in
    fact, the correct behavior.

    This really should be fixed in the transformers library.
    """

    _hf_base_class = Seq2SeqTrainingArguments

    evaluation_strategy: IntervalStrategy = IntervalStrategy.NO
    lr_scheduler_type: SchedulerType = SchedulerType.LINEAR
    logging_strategy: IntervalStrategy = IntervalStrategy.STEPS
    save_strategy: IntervalStrategy = IntervalStrategy.STEPS
    optim: OptimizerNames = OptimizerNames.ADAMW_HF
    hub_strategy: HubStrategy = HubStrategy.EVERY_SAVE



import re
import os
import glob
from enum import Enum
import pandas as pd
import dask.bag as db


_english_punctuation = ".?!"


def _load_from_filename(fn):
    # Load the UTF8 encoded file "fn" and return it as a string.
    with open(fn, encoding='utf-8', mode='r') as f:
        return os.path.basename(fn), f.read()


def _split_sentences(folio):
    # Split a folio into sentences along regular punctuation. Only works for English.
    return [sent.strip()
            for sent in re
                .sub(f"[{_english_punctuation}]", lambda match: match.group() + '\u2406', folio)
                .split("\u2406")]


class CorpusSplitType(Enum):
    """Used to specify how to split the corpus files into individual Dask bag elements."""
    BY_FOLIO = 1
    BY_SECTION = 2
    BY_SENTENCE = 3


class CorpusLoader(object):
    """This is the base class for corpus loader objects. It should not be created by itself.

    Attributes:
        data_glob: A glob for the location of the data after the prefix that locates the repo. The prefix is specified
            in the constructor. For example: OpenPecha/P000001/*.txt.
        exclusions (set): Filenames to exclude from the data glob when loading the dataset. Use for known bad files in
            a particular corpus. NB: this is not a glob.
    """

    _df_column_names, _df_meta = None, None
    data_glob = ''
    glob_exclusions = set()

    def __init__(self,
                 glob_prefix=None,
                 glob_override=None):
        """Constructor for a corpus loader.

        Args:
            glob_prefix (optional): The prefix to the current path for the location of the repo that contains this
                corpus. It will be concatenated with data_glob to get the full path. Defaults to the value of the
                CAI_DATA_BASE_PATH environment variable
            glob_override (optional): Overrides the entire file path glob to specify the location of the corpus. For
                example: ../../my_data/OpenPecha/P000001/*.txt
        """

        super().__init__()
        if glob_prefix is None:
            glob_prefix = os.environ['CAI_DATA_BASE_PATH']
        if glob_override is not None:
            self.data_glob = glob_override
        else:
            self.data_glob = os.path.join(glob_prefix, self.data_glob)

    def _process_bag(self, bag, locators):
        # Apply whatever map functions or other processing to the Dask bag after loading the files and before splitting
        #   the corpus. To be implemented by the child corpora loaders.
        raise NotImplementedError()

    def _make_bag(self, locators):
        # Make a bag, with or without locators, for final loaded outputs.
        bag = db \
            .from_sequence(filter(
                lambda fn: os.path.basename(fn) not in self.glob_exclusions,
                glob.glob(self.data_glob))) \
            .map(_load_from_filename)
        return self._process_bag(bag, locators=locators)

    def to_bag(self, split_type=CorpusSplitType.BY_FOLIO):
        """Return a bag of segmented text, split in the specified way.

        Args:
            split_type (CorpusSplitType): How to split the text into bag elements. Can be BY_FOLIO, BY_SECTION and
                BY_SENTENCE. Defaults to BY_FOLIO. Tibetan can't be split BY_SENTENCE, while English can't be split
                BY_SECTION.

        Returns:
            dask.bag: The Dask bag that contains the processed corpus.
        """

        bag = self._make_bag(locators=False)
        if split_type == CorpusSplitType.BY_SECTION:
            return bag \
                .str.split(' ') \
                .flatten() \
                .str.replace('\n', '')
        elif split_type == CorpusSplitType.BY_FOLIO:
            return bag
        elif split_type == CorpusSplitType.BY_SENTENCE:
            return bag \
                .map(_split_sentences) \
                .flatten()
        else:
            raise ValueError("Wrong split_type: " + str(split_type))

    @property
    def dataframe(self):
        """dask.dataframe: The Dask dataframe that contains the processed corpus and entry metadata."""
        if self._df_meta is not None:
            return self \
                ._make_bag(locators=True) \
                .to_dataframe(meta=pd.DataFrame({
                    col_name: sample_data
                    for col_name, sample_data in zip(self._df_column_names, self._df_meta)}))
        else:
            return self \
                ._make_bag(locators=True) \
                .to_dataframe(columns=self._df_column_names)



import logging

import re
from bs4.element import Comment, NavigableString, Tag, XMLProcessingInstruction
from bs4 import BeautifulSoup, SoupStrainer
from .corpus_loader import CorpusLoader

logger = logging.getLogger(__name__)

def _treat_tags(soup, tag_treatments):
    # Apply the tag_treatments dictionary to the XML soup. Return the soup so that it can be fed to other functions.
    for tag_name, tag_treatment in tag_treatments.items():
        for elem in soup.find_all(tag_name):
            if tag_treatment == 'decompose':
                elem.decompose()
            elif tag_treatment == 'unwrap':
                elem.unwrap()
    return soup


def _treat_divs_and_refs(soup, toh_key, num_toh_keys, refs_to_strip):
    # Decompose refs that are not pointing to the passed in Tohoku number and unwrap divs in the XML soup. Return the
    #   soup so that it can be fed to other functions.
    for ref in soup.find_all("ref"):
        if not "xml:id" in ref.attrs:
            ref.decompose()
        elif ref.attrs.get('type', None) in refs_to_strip:
            ref.decompose()
        else:
            if ('key' not in ref.attrs and num_toh_keys > 1) and not set(ref.attrs.keys()) == {'target'}:
                raise Exception(f"No 'key' attribute in ref tag {ref}, but more than one Tohoku number for the text "
                                f"{toh_key}")
            if not ('key' not in ref.attrs or ref.attrs.get('key', None) == toh_key) or 'cRef' not in ref.attrs:
                ref.decompose()
    for div in soup.find_all("div"):
        div.unwrap()
    return soup


def _folio_str_to_page_num(folio_str):
    # Converts F.a.b to a (relative) page number
    splits = folio_str.split('.')
    return 2 * int(splits[1]) + (1 if splits[2] == 'b' else 0)


def _segment_folios(soup, toh_key, volumes):
    # Iterate through the processed XML soup in parsing order and segment it into folios using the remaining ref tags.
    #   Return the segmentation.
    valid_tag_types = {'folio', 'volume'}
    space_re = re.compile(r"\s+")
    segmentation, cur_volume, vol_start_page, cur_ref, cur_text = [], None, None, None, ''
    for elem in soup.next_elements:
        if isinstance(elem, NavigableString):
            cur_text += space_re.sub(' ', elem.string)
        elif isinstance(elem, Tag):
            elem_type = elem.attrs.get('type', 'folio')
            if not (elem.name == 'ref' and elem_type in valid_tag_types):
                raise ValueError(f"Bad tag: name {elem.name}, type {elem.attrs.get('type', 'none')}, tohoku number "
                                 f"{toh_key}")
            segmentation.append((cur_volume, cur_ref, cur_text))
            if elem_type == 'folio':
                ref_page = _folio_str_to_page_num(elem['cRef'])
                if vol_start_page is None:
                    vol_start_page = ref_page
                if cur_volume is None:
                    if len(volumes) > 1:
                        raise ValueError(f"In {toh_key}, no volume specified in start of doc but more than one volume "
                                          "in metadata.")
                    cur_volume = next(iter(volumes.keys()))
                if not cur_volume in volumes:
                    raise ValueError(f"In {toh_key}, the current volume was {cur_volume} but is not in the volume "
                                     f"metadata ({sorted(list(volumes.keys()))})")
                cur_text, cur_ref = '', ref_page - vol_start_page + volumes[cur_volume]['start_page'] - 1
                if cur_ref > volumes[cur_volume]['end_page']:
                    logger.warning(f"In {toh_key}, the current page was {cur_ref} which is larger than the end page "
                                   f"{volumes[cur_volume]['end_page']} for volume {cur_volume}, "
                                   f"xml:id={elem['xml:id']}")
            elif elem_type == 'volume':
                cur_volume = elem['cRef']
                if not (cur_volume[0] == 'V' and cur_volume[1:].isdigit()):
                    raise ValueError(f"Badly formatted volume number {cur_volume} in tag name {elem.name}, type "
                                     f"{elem.attrs.get('type', 'none')}, tohoku number {toh_key}")
                cur_volume = int(cur_volume[1:])
                vol_start_page = None
        elif isinstance(elem, Comment):
            pass
        elif isinstance(elem, XMLProcessingInstruction):
            pass
        else:
            raise ValueError(f"Bad tag {elem}")
    segmentation.append((cur_volume, cur_ref, cur_text))
    return segmentation


def _process_fstr(args):
    # Parse the XML given in args and return a segmentation into folios. Args have to be a tuple with the following
    #   contents:
    #
    #   filename, Tohoku number, number of Tohoku numbers in filename, contents of the filename as a string,
    #   a dictionary of tag treatments, a list of ref tag types to strip
    text_name, toh_key, num_toh_keys, volumes, fstr, tag_treatments, refs_to_strip = args
    soup = BeautifulSoup(
        fstr,
        'xml',
        parse_only=SoupStrainer("div", type="translation"))
    return [
        (text_name, toh_key, volume, ref, text)
        for volume, ref, text in
            _segment_folios(
                _treat_divs_and_refs(
                    _treat_tags(
                        soup,
                        tag_treatments
                    ),
                    toh_key,
                    num_toh_keys,
                    refs_to_strip
                ),
                toh_key,
                volumes
            )
        ]


def _fn_to_tohs(args):
    # Extract the list of Tohoku numbers from the metadata sourceDesc tag
    #
    # Args has to be a tuple of (filename, contents of the filename as a string). The function returns a list of
    #   tuples, one tuple per Tohoku number in the text, of the form:
    #
    #   [(filename, Tohoku number, count of Tohoku numbers in text, contents of the filename as a string)]
    fn, fstr = args
    soup = BeautifulSoup(
        fstr,
        'xml',
        parse_only=SoupStrainer("sourceDesc")
    )
    toh_metadata = []
    for source in soup.find_all("bibl"):
        volumes = {}
        for volume in source.find_all("volume"):
            volumes[int(volume['number'])] = {
                "start_page": int(volume['start-page']),
                "end_page": int(volume['end-page'])
            }
        toh_metadata.append((source['key'], volumes))
    return [(fn, toh_key, len(toh_metadata), volumes, fstr) for toh_key, volumes in toh_metadata]


def _recto_verso_paginate(args):
    text_name, toh_key, volume, ref, text = args
    p = 0 if ref is None else int(ref)
    ref = f"F.{(p - 1) // 2 + 1}.{'a' if (p % 2) == 0 else 'b'}"
    return (text_name, toh_key, volume, ref, text)


class TeiLoader(CorpusLoader):
    """Corpus loader for the 84,000 TEI translation files. Source repo is https://github.com/84000/all-data.

    Attributes:
        data_glob: A glob for the location of the data after the prefix that locates the repo. The prefix is specified
            in the constructor. Defaults to: 84000/data-tei/translations/{corpus}/translations, you have to specify
            corpus in the constructor, currently can be kangyur or tengyur.
        glob_exclusions (set): Filenames in the corpus to ignore. Defaults to 5 files in the Kangyur, one of which has
            the word OLD in its Tohoku number and the others have no <ref> tags.
        ns: The XML namespace for the TEI tags. Defaults to http://www.tei-c.org/ns/1.0.
        tag_treatments (dictionary): Mapping of XML tags with instructions on how to treat them. There are currently
            two possible tag treatments: unwrap and decompose. The meaning of the operations is as in Beautiful Soup 4.
            NB: The treatments have been carefully set from several days of poring over the dataset. Think twice before
            tinkering with this.
        ref_types_to_strip (set): Reference tags usually (but not always) come with a "type" attribute. This is the
            list of the values of this "type" attribute to remove from the XML. Defaults to bampo, sanskrit and volume,
            all lowercase.
    """

    # Additional columns like volume_number are added after parent class generates _df_column_names
    _df_column_names = ["filename", "tohoku_number", "volume_number", "location", "text"]
    _df_meta = [['a'], ['a'], ['a'], ['a'], ['a']]
    _df_final_columns = ["filename", "tohoku_number", "volume_number", "location", "text"]
    _recto_verso_pagination = False

    data_glob = "raw_datasets/84000-translations-tei/translations/{corpus}/translations/*.xml"
    glob_exclusions = {
        '079-008_toh381OLD-the_emergence_from_samputa%20(20190216).xml',
        '040-005_toh54-the_chapter_on_the_complete_approach.xml',
        '043-006_toh69-arousing_determination.xml',
        '043-006_toh69-the_sutra_which_incites_resolve.xml',
        '057-010_toh151-the_sutra_of_the_questions_of_pratibhanamati.xml',
        '061-012_toh192-the_prophecy_of_ksemavati.xml',
        '066-007_toh249-the_sutra_teaching_four_qualities.xml',
        '067-005_toh266-the_sutra_of_the_flower_collection.xml',
        '077-002_toh361-a_summary_explanation_of_the_initiation.xml',
        '094-006_toh729,1001-the_incantation_of_tara.xml',
        '094-007_toh730,1002-the_incantation_taras_own_promise.xml',
        '048-001_toh101-upholding_the_roots_of_virtue_backup.xml',
        '048-001_toh101-obtaining_the_roots_of_virtue.xml',
        '096-062_toh813,1098-a_prayer_from_destroyer_of_the_great_chiliocosm_cf_toh_558.xml',
        '057-001_toh142-entering_into_non-conceptuality.xml',
        '058-006_toh157-the_sutra_of_the_questions_of_the_kinnara_king_druma.xml',
        '059-008_toh165-the_sutra_of_the_questions_of_ksemamkara.xml',
        '060-007_toh178-the_sutra_of_the_teaching_on_the_aids_to_enlightenment.xml',
        '060-007_toh178-teaching_on_the_aids_to_enlightenment.xml',
        '061-001_toh181-the_sutra_teaching_the_five_perfections.xml',
        '061-016_toh196-the_sutra_on_the_residence_of_manjusri.xml',
        '061-018_toh198-the_sutra_on_maitreyas_setting_forth.xml',
        '062-003_toh203-the_sutra_of_the_seal_of_the_dharma.xml',
        '062-007_toh207-the_sutra_on_the_elephants_exertions.xml',
        '062-007_toh207-the_the_strength_of_the_elephant.xml',
        '062-011_toh211-the_sutra_teaching_the_beginnings_and_the_divisions_of_dependent_arising.xml',
        '066-002_toh244-the_sutra_of_the_way_of_the_dharma.xml',
        '066-003_toh245-the_sections_of_dharmas.xml',
        '066-003_toh245-the_sutra_of_the_collection_of_dharmas.xml',
        '066-005_toh247-distinguishing_between_dharmas_and_their_reality.xml',
        '066-005_toh247-the_sutra_on_how_to_distinguish_dharma_and_worldly_pursuits.xml',
        '066-010_toh252-the_sutra_on_accomplishing_sets_of_four_things.xml',
        '068-002_toh268-the_s%C5%ABtra_of_king_of_the_inconceivable.xml',
        '068-004_toh270,512,852-the_sutra_of_the_seven_buddhas.xml',
        '068-005_toh271-the_sutra_of_the_eight_buddhas.xml',
        '068-007_toh273,511,853-the_sutra_of_the_twelve_buddhas.xml',
        '068-012_toh278-the_sutra_on_the_eightfold_auspiciousnesses.xml',
        '068-019_toh285-the_dedication_perfecting_all_intentions.xml',
        '072-017_toh317-the_dharma_discourse_on_the_ascertainment_of_the_meaning.xml',
        '072-025_toh325-verses_on_the_naga_king_bheri.xml',
        '075-003_toh345-the_account_of_noble_deeds_concerning_a_sow.xml',
        '088-008_toh514,854-the_incantation_the_essence_of_the_buddhas__a_dharma_discourse.xml',
        '088-009_toh515,855-the_incantation_the_essence_of_the_buddhas.xml',
        '088-016_toh522,848-the_dharani_of_the_torch_of_gnosis.xml',
        '042-002_toh61-the_sutra_of_purnas_questions.xml',
        '042-003_toh62-the_sutra_of_rastrapalas_questions.xml',
        '091-072_toh674,849-the_aparimitayurjnana_sutra.xml',
        '091-072_toh674,849-the_sutra_of_immeasurable_life_and_gnosis.xml',
        '091-074_toh676,850-the_incantation_the_essence_of_immeasurable_life_and_gnosis.xml',
        '091-077_toh679,851-the_incantation_immeasurable_praise_of_good_qualities.xml',
        '043-007_toh70-question_of_subahu.xml',
        '043-007_toh70-the_sutra_of_subahus_questions.xml',
        '094-002_toh725,909-the_incantation_mother_of_avalokitesvara.xml',
        '094-008_toh731-the_sutra_of_tara_who_protects_from_the_eight_fearful_things.xml',
        '094-013_toh736,995-the_incantation_of_parnasabari.xml',
        '100-001_toh846-the_three_part_invitation_tantra.xml',
        '100-002_toh846a-the_threefold_ritual.xml'}
    ns = "http://www.tei-c.org/ns/1.0"
    tag_treatments = {
        'term': 'unwrap',
        'lg': 'unwrap',
        'lb': 'decompose',
        'milestone': 'decompose',
        'foreign': 'unwrap',
        'distinct': 'unwrap',
        'emph': 'unwrap',
        'title': 'unwrap',
        'head': 'unwrap',
        'q': 'unwrap',
        'l': 'unwrap',
        'trailer': 'unwrap',
        'note': 'decompose',
        'p': 'unwrap',
        'mantra': 'unwrap',
        'list': 'unwrap',
        'item': 'unwrap',
        'label': 'decompose',
        'canonDef': 'unwrap',
        'media': 'decompose',
        'fix': 'decompose'}
    ref_types_to_strip = {'bampo', 'sanskrit'}

    def __init__(self,
                 tei_corpus,
                 glob_prefix=None,
                 glob_override=None):
        """Constructor for a corpus loader for the 84,000 TEI translation files.

        Args:
            glob_prefix (optional): The prefix to the current path for the location of the repo that contains this
                corpus. It will be concatenated with data_glob to get the full path. For example: ../../my_data
            tei_corpus: Which of the 84,000 translated corpora to load. It is used to form the directory that the XML
                files are in: 84000/data-tei/translations/{tei_corpus}/translations. Currently, can be kangyur or
                tengyur.
            glob_override (optional): Overrides the entire file path glob to specify the location of the corpus. For
                example: ../../my_data/84000/data-tei/translations/kangyur/translations/*.xml
        """

        self.data_glob = self.data_glob.format(corpus=tei_corpus)
        super().__init__(
            glob_prefix=glob_prefix,
            glob_override=glob_override)

    def recto_verso_pagination(self, recto_verso=True):
        """Set the pagination to F.[page].[a=recto/b=version] style.

        Args:
            recto_verso: Use recto-verso pagination if True. Defaults to True.

        Returns:
            The corpus loader object so that methods can be chained in the functional style.
        """

        self._recto_verso_pagination = recto_verso
        return self

    def _process_bag(self, bag, locators):
        # Prepares a bag, with or without locators as indicated. Adds the Tohoku numbers, applies folio segmentation,
        #   strips markup, removes spaces, and removes locators if requested.
        bag = bag \
            .map(_fn_to_tohs) \
            .flatten() \
            .map(lambda args: tuple(list(args) + [self.tag_treatments, self.ref_types_to_strip])) \
            .map(_process_fstr) \
            .flatten()
        if self._recto_verso_pagination:
            bag = bag \
                .map(_recto_verso_paginate)
        if not locators:
            bag = bag.map(lambda args: args[-1])
        return bag

    @property
    def dataframe(self):
        res_df = super().dataframe
        res_df["volume_number"] = res_df["volume_number"].fillna(
            res_df.filename.map(lambda x: x.split('-')[0])) \
            .astype(int)
        res_df = res_df[self._df_final_columns]
        return res_df



def repl_split_commas_or_kill(match, take_second_item=False, test_mode=False):
    # Take a regex match of the form (...,...), or some other brackets, and return one of the items.
    #
    # Test mode returns the bracketed input and the result, for unit testing.
    match = match.group(0)
    if ',' in match:
        if take_second_item:
            if test_mode:
                return f"@{match}|{match.split(',')[1][:-1]}@"
            else:
                return match.split(',')[1][:-1]
        return match.split(',')[0][1:]
    # This is a different comma in unicode but it looks the same in most fonts
    if '‚' in match:
        if take_second_item:
            if test_mode:
                return f"@{match}|{match.split('‚')[1][:-1]}@"
            else:
                return match.split('‚')[1][:-1]
        return match.split('‚')[0][1:]
    return ''



import argparse
import random


def main(args):     # pylint: disable=redefined-outer-name
    # Shuffle two files in parallel. Useful for parallel training data preparation.
    with open(args.input_file + '.bo', 'r', encoding='utf-8') as f_bo, open(args.input_file + '.en', 'r') as f_en:
        bo_lines, en_lines = f_bo.readlines(), f_en.readlines()
    combined = list(zip(bo_lines, en_lines))

    random.shuffle(combined)
    if args.num_lines > 0:
        combined = combined[:args.num_lines]

    with open(args.output_file + '.bo', 'a', encoding='utf-8') as f_bo, open(args.output_file + '.en', 'a') as f_en:
        for bo_line, en_line in combined:
            f_bo.write(bo_line)
            f_en.write(en_line)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-i',
        '--input-file',
        type=str,
        help='Input filename, without the extension')
    parser.add_argument(
        '-o',
        '--output-file',
        type=str,
        help='Output filename, without the extension')
    parser.add_argument(
        '-n',
        '--num-lines',
        type=int,
        default=-1,
        help='Number of lines to output')
    parser.add_argument(
        '--seed',
        type=int,
        default=12345,
        help='Random seed')
    args = parser.parse_args()

    main(args)



import os
import glob
import json
import yaml
import logging

from urllib.error import HTTPError
from tqdm.auto import tqdm
from torch.hub import download_url_to_file, get_dir
from cai_common.defaults import cai_s3


logger = logging.getLogger(__name__)

os_data_base_path = os.environ.get('CAI_DATA_BASE_PATH', None)
if os_data_base_path in {'', '***unset***'}:
    os_data_base_path = None


def _download_file(s3_loc, target_fn):
    try:
        download_url_to_file(s3_loc, target_fn)
    except HTTPError as e:
        logger.error(f"Failed to download {s3_loc} to {target_fn}")
        raise e


def get_local_model_dir(model_name, download_if_missing=True):
    """Load a model from the CAI S3 data registry.
    
    Args:
        model_name (:obj:`string`):
            The model name in the CAI data registry. If it starts with 'model_archive', it is assumed to be a path
            within the data registry. Otherwise, it is assumed to be a champion model inside the champion_models
            directory.
        download_if_missing (:obj:`bool`, `optional`): Download from the CompassionAI S3 repository if missing the local
            repository. This is expected in inference installations. Defaults to True.

    Returns:
        The local directory name you can feed to AutoModel.from_pretrained.
    """
    if model_name.startswith("https://"):
        s3_model_loc = model_name
        model_name = model_name[len(cai_s3):]
        model_dir = os.path.join(get_dir(), model_name)
    elif os.path.exists(model_name):
        model_dir = model_name
        s3_model_loc = None
    else:
        data_base_path = os_data_base_path if os_data_base_path is not None else cai_s3
        if data_base_path.startswith("https://"):
            model_dir = os.path.join(get_dir(), model_name)
            s3_model_loc = os.path.join(data_base_path, model_name)
        else:
            model_dir = os.path.join(data_base_path, model_name)
            s3_model_loc = None

    if s3_model_loc is not None:
        if not s3_model_loc.startswith(cai_s3):
            raise ValueError("Only downloads from the public CAI data registry are allowed")

        os.makedirs(model_dir, exist_ok=True)
        
        manifest_fn = os.path.join(model_dir, "manifest.yaml")
        if not os.path.exists(manifest_fn):
            if not download_if_missing:
                raise FileNotFoundError("Model not found locally and download if missing is set to False.")
            _download_file(f"{s3_model_loc}/manifest.yaml", manifest_fn)
        with open(manifest_fn, 'r') as manifest_f:
            manifest = yaml.safe_load(manifest_f)

        to_download = [
            fn for fn, f_loc in [(fn, os.path.join(model_dir, fn)) for fn in manifest['inference']]
                if not os.path.exists(f_loc)
        ]
        if len(to_download) > 0:
            if not download_if_missing:
                raise FileNotFoundError("Model not found locally and download if missing is set to False.")
            for model_file in tqdm(to_download, desc="Downloading model files"):
                _download_file(f"{s3_model_loc}/{model_file}", os.path.join(model_dir, model_file))

    return model_dir


def get_local_file(file_subpath):
    """Load a file from the CAI S3 data registry. Will download the file if needed.
    
    Args:
        file_subpath (:obj:`string`):
            The subpath of the file within the CAI data registry.

    Returns:
        The local file name.
    """
    if os.path.exists(file_subpath):
        return file_subpath
    file_dir = get_local_model_dir(os.path.dirname(file_subpath))
    return os.path.join(file_dir, os.path.basename(file_subpath))


def get_local_ckpt(model_name, model_dir=False, search_for_ext="bin", download_if_missing=True):
    """Convert the name of the model in the CAI data registry to a local checkpoint path.

    Args:
        model_name (:obj:`string`):
            The model name in the CAI data registry. If it starts with 'model_archive', it is assumed to be a path
            within the data registry. Otherwise, it is assumed to be a champion model inside the champion_models
            directory.

            If model name has no extension, and model_dir is False, it checks if there is only one file with the
            extension specified in search_for_ext in the path the model name resolves to. If there is more than one, it
            crashes, otherwise it returns the path to the unique file with the requested extension.
        model_dir (:obj:`bool`, `optional`): Return the model directory, not a candidate file. Useful for Hugging Face
            local loading using from_pretrained.
        search_for_ext (:obj:`string`, `optional`): What extension to search for. Defaults to 'bin'.
        download_if_missing (:obj:`bool`, `optional`): Download from the CompassionAI S3 repository if missing the local
            repository. This is expected in inference installations. Defaults to True.

    Returns:
        The local directory name you can feed to AutoModel.from_pretrained.
    """

    if os_data_base_path is None and not download_if_missing:
        raise FileNotFoundError("CAI data registry path not set and downloading is switched off")
    if not model_name.startswith('experiments'):
        model_name = os.path.join('champion_models', model_name)

    model_name = get_local_model_dir(model_name)

    if model_dir:
        return model_name
    if os.path.splitext(model_name)[1] == '':
        candidates = glob.glob(os.path.join(model_name, "*." + search_for_ext))
        if len(candidates) == 0:
            raise FileNotFoundError(f"No .{search_for_ext} files found in {model_name}")
        if len(candidates) > 1:
            raise FileExistsError(f"Multiple .{search_for_ext} files in {model_name}, please specify which one to load "
                                   "by appending the .{search_for_ext} filename to the model name")
        model_name = candidates[0]
    return model_name


def get_cai_config(model_name):
    """Load the CompassionAI config for the name of the model in the CAI data registry.

    Args:
        model_name (:obj:`string`):
            The model name in the CAI data registry. Follows the same rules as get_local_ckpt.

    Returns:
        The loaded CompassionAI config JSON.
    """

    cfg_fn = get_local_ckpt(model_name, search_for_ext="config_cai.json")
    if not cfg_fn[-len(".config_cai.json"):] == ".config_cai.json":
        cfg_fn = cfg_fn.split('.')[0] + ".config_cai.json"      # Deliberately take the first dot
    with open(cfg_fn) as f:
        return json.load(f)



import os
from enum import Enum

import pyewts
import dask.bag as db
from cai_common.utils import open_globs


class TibetanEncoding(Enum):
    """Choices of encoding for Tibetan: EWTS (Extended Wylie) or Unicode."""
    UNICODE = 1
    EWTS = 2


class TibetanDict:
    """Dictionary object to translate individual words from Tibetan into English. Supports Extended Wylie and Unicode
    Tibetan encodings.

    Attributes:
        dict_glob (iterable(str) or str): If a string then this is a glob for the location of dictionaries after the
            prefix that locates the repo. If an iterable of strings then each string is treated as a glob and the
            results are unioned to get all the dictionary files.  Defaults to the string
            raw_datasets/tibetan-english-dictionaries/*
        glob_exclusions (iterable(str)): Filenames to exclude from the data glob when loading the dictionaries. Use for
            known bad files. NB: this is not a glob.
        separator (char): The separator character in the dictionary files. To the left of the character is Tibetan in
            Extended Wylie encoding and to the right is the target language. Defaults to '|'.
    """

    _dictionary = {}
    _default_encoding = TibetanEncoding.UNICODE
    dict_glob = "raw_datasets/tibetan-english-dictionaries/*"
    glob_exclusions = set()
    separator = '|'

    def __init__(self,
                 glob_prefix=None,
                 glob_override=None,
                 default_encoding=TibetanEncoding.UNICODE):
        """Constructor for a Tibetan dictionary.

        Args:
            glob_prefix (optional): The prefix to the current path for the location of the repo that contains the
                dictionary files. It will be concatenated with dict_glob to get the full path. For example:
                ../../tibert_data
            glob_override (optional): Overrides the postfix file path glob to specify the location of the corpus inside
                the data registry. For example: raw_datasets/tibetan-english-dictionaries/*
        """

        super().__init__()
        if glob_prefix is None:
            glob_prefix = os.environ['CAI_DATA_BASE_PATH']
        if glob_override is not None:
            self.dict_glob = glob_override
        self.dict_glob = os.path.join(glob_prefix, self.dict_glob)
        self._dict_files = open_globs(self.dict_glob, self.glob_exclusions)
        self._ewts_converter = pyewts.pyewts()
        self._default_encoding = default_encoding
        self._load_dicts(self._dict_files)

    def _process_dict_entry(self, dict_entry):
        # Normalize encoding and translate a single Tibetan word
        tibetan_word, translation = dict_entry.split(self.separator)
        if self.default_encoding == TibetanEncoding.UNICODE:
            tibetan_word = self.EWTS_to_Unicode(tibetan_word)
        elif self.default_encoding == TibetanEncoding.EWTS:
            pass
        else:
            raise ValueError("Bad default_encoding: " + self.default_encoding)
        return (tibetan_word, translation)

    def _load_dicts(self, dict_names):
        # Load all dictionary files in dict_names. The dict_names arg should contain full paths.
        self._dictionary = {}
        _dict_bag = db \
            .read_text(dict_names) \
            .filter(lambda dict_entry: self.separator in dict_entry) \
            .map(self._process_dict_entry) \
            .map(lambda args: (args[0], [t.strip() for t in args[1].split(';')])) \
            .foldby(lambda x: x[0], lambda total, x: total + x[1], [], lambda total1, total2: total1 + total2, [])
        self._dictionary = dict(_dict_bag.compute())

    def __getitem__(self, tibetan):
        """Translate a Unicode Tibetan word into the target language.

        Args:
            tibetan (str or list): Either a single Tibetan word or a list of Tibetan words, encoded according to
                default_encoding.

        Returns:
            List of all dictionary translations for the Tibetan word, or a list of lists if multiple words were
            provided, or None if the word is not found."""
        if isinstance(tibetan, list):
            return [self[tw] for tw in tibetan]
        if tibetan[-1] == '་':
            tibetan = tibetan[:-1]
        return self._dictionary.get(tibetan, None)

    def __len__(self):
        """Get the number of words in the dictionary.

        Returns:
            The number of words in the dictionary."""

        return len(self._dictionary)

    def EWTS_to_Unicode(self, wylie_tibetan):
        """Convert an Extended Wylie encoded Tibetan word into Unicode.

        Args:
            wylie_tibetan: A Tibetan string encoded in Extended Wylie.

        Returns:
            The Unicode encoding of the same Tibetan string."""
        return self._ewts_converter.toUnicode(wylie_tibetan)

    def Unicode_to_EWTS(self, unicode_tibetan):
        """Convert a unicode encoded Tibetan word into Extended Wylie.

        Args:
            wylie_tibetan: A Tibetan string encoded in unicode.

        Returns:
            The Extended Wylie encoding of the same Tibetan string."""
        return self._ewts_converter.toWylie(unicode_tibetan)

    def all_tibetan_words(self):
        """Return all the Tibetan words this dictionary can translate.

        Returns:
            An iterator through all the Tibetan words this dictionary can translate.
        """

        return self._dictionary.keys()

    def keys(self):
        """Synonym for all_tibetan_words."""

        return self.all_tibetan_words()

    def items(self):
        """Return all the Tibetan words this dictionary can translate with their English definitions.

        Returns:
            An iterator through all the Tibetan words this dictionary can translate with their English definitions.
        """

        return self._dictionary.items()

    @property
    def dict_files(self):
        """Read-only property for the actual files used for the dictionaries."""

        return self._dict_files

    @property
    def default_encoding(self):
        """TibetanEncoding: Which encoding for Tibetan to use when translating. Can be UNICODE or EWTS. Changing this
            causes the dictionary to recalculate. Defaults to UNICODE.
        """

        return self._default_encoding

    @default_encoding.setter
    def default_encoding(self, encoding):
        self._default_encoding = encoding
        self._load_dicts(self._dict_files)



import re
from cai_common.dict import tibetan_digits, translate_tibetan_number

from .corpus_loader import CorpusLoader
from .utils import repl_split_commas_or_kill


def _apply_with_locators(args, func, locators):
    # Apply a function to a bag element, taking the presence or absence of locators into account.
    if locators:
        args = list(args)
        return tuple(args[:-1] + [func(args[-1])])
    else:
        return func(args)


class OpenPechaLoader(CorpusLoader):
    """This is the base class for corpus loaders of material from the OpenPecha repositories. It should not be created
    by itself.

    Attributes:
        data_glob: A glob for the location of the data after the prefix that locates the repo. The prefix is specified
            in the constructor. For example: OpenPecha/P000001/*.txt
    """

    _space_re = re.compile(r"\s+")
    _folio_marker_re = None
    _apply_markup = True
    _replace_with_suggested = False
    _clean_bad_chars = True
    _remove_section_headings = True
    _remove_new_lines = False

    def __init__(self,
                 glob_prefix=None,
                 glob_override=None,
                 test_mode=False):
        """Constructor for a corpus loader from an OpenPecha repository. The default settings are to apply markup, not
        to replace with suggestions in the markup, to clean up bad characters, and to remove section headings.

        Args:
            glob_prefix (optional): The prefix to the current path for the location of the repo that contains this
                corpus. It will be concatenated with data_glob to get the full path. Defaults to the value of the
                CAI_DATA_BASE_PATH environment variable
            glob_override (optional): Overrides the entire file path glob to specify the location of the corpus. For
                example: ../../my_data/OpenPecha/P000001/*.txt
            test_mode: When applying markup, retain the original and the markup being applied. Formatted as @(x,y)|res@
                where res is either x or y depending on the setting of replace_with_suggested given in apply_markup.
                Defaults to False, intended for testing the markup application.
        """

        super().__init__(
            glob_prefix=glob_prefix,
            glob_override=glob_override)
        self.test_mode = test_mode

    def apply_markup(self, apply=True, replace_with_suggested=False):
        """Set whether to apply the markup for suggested replacements within the corpus.

        Args:
            apply: Markup is treated if apply=True. If apply=False then the markup is not touched at all and the
                brackets are left as is. Defaults to True.
            replace_with_suggested: If True then the suggested replacements in the markup are performed. This means
                that elements like (x,y) become y. Defaults to False.

        Returns:
            The corpus loader object so that methods can be chained in the functional style.
        """

        self._apply_markup = apply
        self._replace_with_suggested = replace_with_suggested
        return self

    def clean_bad_chars(self, clean=True):
        """Set whether to remove bad characters from the corpus. Each corpus has to specify its own bad character set.

        Args:
            clean: Remove bad characters if True. Defaults to True.

        Returns:
            The corpus loader object so that methods can be chained in the functional style.
        """

        self._clean_bad_chars = clean
        return self

    def remove_section_headings(self, remove=True):
        """Set whether to remove various headings from the corpus, such as Tohoku catalog numbers or folio numbers.

        Args:
            clean: Remove headings if True. Defaults to True.

        Returns:
            The corpus loader object so that methods can be chained in the functional style.
        """

        self._remove_section_headings = remove
        return self

    def remove_new_lines(self, remove=True):
        """Set whether to remove newline characters from the corpus. This will connect sections split by line breaks
            but not by folio breaks.

        Args:
            clean: Remove newlines if True. Defaults to False.

        Returns:
            The corpus loader object so that methods can be chained in the functional style.
        """

        self._remove_new_lines = remove
        return self

    def _format_folio_locator(self, match):
        raise NotImplementedError()

    def _get_volume_num_from_fn(self, fn):
        return int(fn.split("_")[0])

    def _segment_folios(self, args):
        # Segment folios using the folio marker regular expression by stepping through successive folio marker matches
        #   and taking the text in between. Also converts the Tibetan numerals for page markers into 84,000-style folio
        #   locators of the form F.page.a(verso)/b(recto).
        fn, volume_str = args
        volume_num = self._get_volume_num_from_fn(fn)
        res, cur_ref, last_loc = [], None, None
        if self._folio_marker_re is None:
            raise NotImplementedError()
        for match in self._folio_marker_re.finditer(volume_str):
            cur_loc_start, cur_loc_end = match.span()
            if last_loc is not None:
                res.append((fn, volume_num, cur_ref, volume_str[last_loc:cur_loc_start]))
            last_loc = cur_loc_end
            cur_ref = self._format_folio_locator(match)
        res.append((fn, volume_num, cur_ref, volume_str[last_loc:]))
        return res

    def _process_bag(self, bag, locators):
        # Prepares a bag, with or without locators as indicated. Applies folio segmentation, removes spaces, and
        #   removes locators if requested.
        bag = bag \
            .map(self._segment_folios) \
            .flatten() \
            .map(
                lambda args: _apply_with_locators(
                    args,
                    lambda x:
                        self._space_re.sub(' ', x.replace('\n', '') if self._remove_new_lines else x).strip(), True))
        if not locators:
            bag = bag \
                .map(lambda args: args[-1]) \
                .filter(lambda x: not x == '')
        return bag


class OldKangyurLoader(OpenPechaLoader):
    """Esukhia Derge Kangyur corpus loader. This is for the old version of the eKangyur, archived in our data registry
        under OpenPecha-kangyur-old.

    Attributes:
        data_glob: A glob for the location of the data after the prefix that locates the repo. The prefix is specified
            in the constructor. Defaults to: raw_datasets/OpenPecha-kangyur-old/*.txt
    """

    _df_column_names = ["filename", "volume_number", "location", "text"]
    _df_meta = [['a'], ['a'], ['a'], ['a']]
    _folio_marker_re = re.compile(rf"\(([{''.join(tibetan_digits)}]+)([ནབ])་\)")
    data_glob = "raw_datasets/OpenPecha-kangyur-old/*.txt"

    def __init__(self,
                 glob_prefix=None,
                 glob_override=None,
                 test_mode=False):
        """Constructor for the Kangyur loader from the OpenPecha repository. The default settings are to apply markup,
        not to replace with suggestions in the markup, to clean up bad characters, and to remove section headings.

        Args:
            glob_prefix (optional): The prefix to the current path for the location of the repo that contains this
                corpus. It will be concatenated with data_glob to get the full path. Defaults to the value of the
                CAI_DATA_BASE_PATH environment variable
            glob_override (optional): Overrides the entire file path glob to specify the location of the corpus. For
                example: ../../my_data/OpenPecha/P000001/*.txt
            test_mode: When applying markup, retain the original and the markup being applied. Formatted as @(x,y)|res@
                where res is either x or y depending on the setting of replace_with_suggested given in apply_markup.
                Defaults to False, intended for testing the markup application.
        """

        super().__init__(
            glob_prefix=glob_prefix,
            glob_override=glob_override,
            test_mode=test_mode)

    def _format_folio_locator(self, match):
        # Format the result of the folio marker regular expression (self._folio_marker_re) to be in the standard 84,000
        #   form F.page.a(verso)/b(recto).
        return f"F.{translate_tibetan_number(match.group(1))}.{'a' if match.group(2) == 'ན' else 'b'}"

    def _preprocess_bag(self, bag, _locators):
        # Child classes can inherit this to do their own processing before this class mangles the text
        return bag

    def _process_bag(self, bag, locators):
        # Prepares a bag, with or without locators as indicated. Applies folio segmentation, removes spaces, and
        #   removes locators if requested.
        #
        # Basic strategy:
        #    0. First segment into folios if required
        #    1. Find all round and curly braces (no curly braces in the Kangyur but inspired by the Tengyur we still do
        #       it)
        #        a. If there is a comma in the bracketed text, split along comma and take first element (uncorrected)
        #        b. If no comma then just kill entire bracketed text (remove comments and mark-up)
        #    2. Remove square brackets and leave text inside them (presumed interpolation for damage etc)
        #    3. Remove unmatched commas, brackets, stars, the BOM, etc (leaves small number of bad morphemes that are
        #       very hard to clean)
        bag = super()._process_bag(bag, locators)
        bag = self._preprocess_bag(bag, locators)
        if self._apply_markup:
            # Have to repeat twice because some brackets are nested to depth 2
            bag = bag.map(lambda args: _apply_with_locators(
                args,
                lambda x: re.sub(
                    r"[({][^(){}]*?[})]",
                    lambda y: repl_split_commas_or_kill(y, self._replace_with_suggested, self.test_mode),
                    x),
                locators))
            if not self.test_mode:
                bag = bag.map(lambda args: _apply_with_locators(
                    args,
                    lambda x: re.sub(
                        r"[({][^(){}]*?[})]",
                        lambda y: repl_split_commas_or_kill(y, self._replace_with_suggested, self.test_mode),
                        x),
                    locators))
        if self._clean_bad_chars:
            # There are opening/closing brackets with no matching brackets
            bad_chars_to_remove = "*`p}1-9\ufeff�‘’"
            bag = bag.map(lambda args: _apply_with_locators(
                args,
                lambda x: re.sub(rf"[\[\]{bad_chars_to_remove}]", '', x),
                locators))
        return bag


class KangyurLoader(OldKangyurLoader):
    """Esukhia Derge Kangyur corpus loader. This is for the new version of the eKangyur, stored in our data registry
        under OpenPecha-kangyur.

    Note that it is currently not paginated.

    Attributes:
        data_glob: A glob for the location of the data after the prefix that locates the repo. The prefix is specified
            in the constructor. Defaults to: raw_datasets/OpenPecha-kangyur/*.txt
        base_folio_number: The number of the first page in every volume. Defaults to 3, which is what is currently
            in the OpenPecha dataset. Used to rebase the pagination so that every volume starts from page 1, just like
            in the eKangyur pagination used by 84,000.
        recto_verso_pagination: Reproduce the F.x.(a|b) pagination format. Defaults to False.
    """

    _df_column_names = ["filename", "volume_number", "location", "text"]
    _df_meta = [['a'], ['a'], ['a'], ['a']]
    _folio_marker_re = re.compile(r" i-\d+")
    _uuid_re = re.compile(r"OpenPechaUUID:[\w{6,6}]")
    _remove_uuids = True
    data_glob = "raw_datasets/OpenPecha-kangyur/*.txt"
    base_folio_number = 3
    recto_verso_pagination = False

    def __init__(self,
                 glob_prefix=None,
                 glob_override=None,
                 test_mode=False):
        """Constructor for the Kangyur loader from the OpenPecha repository. The default settings are to remove the
            OpenPechaUUID's.

        Args:
            glob_prefix (optional): The prefix to the current path for the location of the repo that contains this
                corpus. It will be concatenated with data_glob to get the full path. Defaults to the value of the
                CAI_DATA_BASE_PATH environment variable
            glob_override (optional): Overrides the entire file path glob to specify the location of the corpus. For
                example: ../../my_data/OpenPecha/P000001/*.txt
        """

        super().__init__(
            glob_prefix=glob_prefix,
            glob_override=glob_override,
            test_mode=False)

    def remove_uuids(self, remove=True):
        """Set whether to remove OpenPechaUUIDs from the corpus.

        Args:
            clean: Remove uuids if True. Defaults to True.

        Returns:
            The corpus loader object so that methods can be chained in the functional style.
        """

        self._remove_uuids = remove
        return self

    def _format_folio_locator(self, match):
        # Currently unpaginated. This is because the pagination in the OpenPecha repo doesn't have the recto/verso
        #   tags preserved, and is instead matched to the names of the image files.
        page = int(match[0].split('-')[1])
        if self.recto_verso_pagination:
            folio_num = (page - self.base_folio_number) // 2 + 1
            recto_verso = 'a' if (page % 2) == 1 else 'b'
            return f"F.{folio_num}.{recto_verso}"
        else:
            return str(page - self.base_folio_number)

    def _get_volume_num_from_fn(self, fn):
        return int(fn.split("_")[0][1:])

    def _preprocess_bag(self, bag, locators):
        # Removes the OpenPechaUUIDs if requested to remove section headings
        if self._remove_uuids:
            bag = bag.map(lambda args: _apply_with_locators(
                args,
                lambda x: re.sub(r"OpenPechaUUID:\w{6,6}", '', x),
                locators))
        return bag


class TengyurLoader(OpenPechaLoader):
    """Esukhia Derge Tengyur corpus loader. Source repo is https://github.com/Esukhia/derge-tengyur.

    Attributes:
        data_glob: A glob for the location of the data after the prefix that locates the repo. The prefix is specified
            in the constructor. Defaults to: derge-tengyur/text/*.txt.
    """

    _df_column_names = ["filename", "volume_number", "location", "text"]
    _folio_marker_re = re.compile(r"\[([\d]+)([ab])\]")
    data_glob = "raw_datasets/Esukhia-derge-tengyur/text/*.txt"

    def __init__(self,
                 glob_prefix=None,
                 glob_override=None,
                 test_mode=False):
        """Constructor for the Tengyur loader from the Esukhia repository. The default settings are to apply markup,
        not to replace with suggestions in the markup, to clean up bad characters, and to remove section headings.

        Args:
            glob_prefix (optional): The prefix to the current path for the location of the repo that contains this
                corpus. It will be concatenated with data_glob to get the full path. Defaults to the value of the
                CAI_DATA_BASE_PATH environment variable
            glob_override (optional): Overrides the entire file path glob to specify the location of the corpus. For
                example: ../../my_data/derge-tengyur/text/*.txt
            test_mode: When applying markup, retain the original and the markup being applied. Formatted as @(x,y)|res@
                where res is either x or y depending on the setting of replace_with_suggested given in apply_markup.
                Defaults to False, intended for testing the markup application.
        """

        super().__init__(
            glob_prefix=glob_prefix,
            glob_override=glob_override,
            test_mode=test_mode)

    def _format_folio_locator(self, match):
        # Format the result of the folio marker regular expression (self._folio_marker_re) to be in the standard 84,000
        #   form F.page.a(verso)/b(recto).
        return f"F.{match.group(1)}.{match.group(2)}"

    def _process_bag(self, bag, locators):
        # Prepares a bag, with or without locators as indicated. Applies folio segmentation, removes spaces, and
        #   removes locators if requested.
        #
        # Basic strategy:
        #    1. Find all round and curly braces
        #        a. If there is a comma in the bracketed text, split along comma and take first element (uncorrected)
        #        b. If no comma then just kill entire bracketed text (remove comments and mark-up)
        #    2. Remove text of the form NNNab.NNN where N are numebrs inside square brackets (section/folio/work/etc
        #       numbering)
        #    3. Remove square brackets and leave text inside them (presumed interpolation for damage etc)
        #    4. Remove hashtags (mark-up and comments)
        #    5. Remove unmatched commas, brackets, stars, and the BOM (leaves small number of bad morphemes that are
        #       very hard to clean)
        bag = super()._process_bag(bag, locators)
        if self._apply_markup:
            # Have to repeat twice because some brackets are nested to depth 2
            bag = bag.map(lambda args: _apply_with_locators(
                args,
                lambda x: re.sub(
                    r"[({][^(){}]*?[})]",
                    lambda y: repl_split_commas_or_kill(y, self._replace_with_suggested, self.test_mode),
                    x),
                locators))
            if not self.test_mode:
                bag = bag.map(lambda args: _apply_with_locators(
                    args,
                    lambda x: re.sub(
                        r"[({][^(){}]*?[})]",
                        lambda y: repl_split_commas_or_kill(y, self._replace_with_suggested, self.test_mode),
                        x),
                    locators))
        if self._remove_section_headings:
            bag = bag.map(lambda args: _apply_with_locators(
                args,
                lambda x: re.sub(r"\[[0-9 ]+[ab.0-9]+\]", '', x),
                locators))
        if self._clean_bad_chars:
            # There are opening/closing brackets with no matching brackets
            bad_chars_to_remove = "\ufeff\x07\t+./0-9:;<=>A-Za-z{×"
            bag = bag.map(lambda args: _apply_with_locators(
                args,
                lambda x: re.sub(fr"[\[\]#\(\),*{bad_chars_to_remove}]", '', x),
                locators))
        return bag
